

import json, requests
from datetime import datetime, timedelta, timezone
from tqdm import tqdm as tq
import os, sys, random, pkgutil, math
from rich.progress import Progress
from loguru import logger as logger_
import pickle

from custom_decisions import *
from custom_transcription import *


import marlin_brahma.world.population as pop
import marlin_brahma.fitness.performance as performance
from marlin_brahma.evo.brahma_evo import *

#----------------------------------------------------------------------------
#                           IDENT GAME                                    #
#----------------------------------------------------------------------------



class GameMetrics(object):
    """GameMetrics | Handles all metrics for game and evolutionary performance. Simple returns for plotting functionality.

    :param object: _description_
    :type object: _type_
    """
    def __init__(self, game_id = "", dump_path = ""):
        
        self.game_id = game_id
        # generation data
        self.by_generation = []
        
        self.energy_tracker = {}
        # optimisation overview -> first view of framework running
        self.overview = {}
        # list of all saved bots (with gen number)
        self.saved_bots = []
        # p/l for all bots when saved
        self.results = {}
        
        self.dump_path = dump_path
        
    def update_overview(self, overview = {}):
        self.overview = overview
    
    
        
    

    
    def update_saved_bots(self, bot_list=[], results={}):
        
        
        for bot_id in bot_list:
            bot_root = '_'.join(bot_id.split('_')[:-1])
            self.saved_bots.append(bot_id)
            self.results[bot_id] = results[bot_root]
        
        
    def update_gen(self, data = {}):
       
        gen_number = data['generation_number']
        tmp_results = {}
        for bot_id, result in data['results'].items():
            bot_id_ = f'{bot_id}_{gen_number}'
            tmp_results[bot_id_] = result
        
        data['results'] = tmp_results
        self.by_generation.append(data)
    
    def update_energy(self, tag, generation, data):
        if tag not in self.energy_tracker:
            self.energy_tracker[tag] = {}
        if generation not in self.energy_tracker[tag]:
            self.energy_tracker[tag][generation] = []
        
        self.energy_tracker[tag][generation].append(data)
    
    def dump(self, winners=[], generation = 0):
        
        self.by_generation[generation]['overview'] = {
            'saved_bots' : self.saved_bots,
            'results' : self.results
        }
        
        # overview data
        self.by_generation[generation]['parms'] = self.overview
        
        with open(f'{self.dump_path}/learn_out/gen_data_{self.game_id}.json', 'w') as fp:
            json.dump(self.by_generation,fp)
    
        with open(f'{self.dump_path}/energies/energy_data_{self.game_id}.json', 'w') as fp:
            json.dump(self.energy_tracker,fp)
        
        
        
    
        
    
class IdentGame(object):
    """IdentGame _summary_

    :param object: _description_
    :type object: _type_
    """
    def __init__(self, population = None,performance = None,data_feed = None, derived_data = None, multiple_derived_data = {},game_id = "", game_parms = {}, label_data=[], dmdb_flag = False, init_expression_load = False, dump_path="", bot_path=""):
        """__init__ _summary_

        :param data_manager: _description_, defaults to None
        :type data_manager: _type_, optional
        :param game_id: _description_, defaults to ""
        :type game_id: str, optional
        :param game_parms: _description_, defaults to {}
        :type game_parms: dict, optional
        :param label_data: _description_, defaults to []
        :type label_data: list, optional
        """
        
         # ---- DUMP PATHS ---
        self.dump_path = dump_path
        if self.dump_path == "":
            return None
        
        self.bot_path = bot_path
        if self.bot_path == "":
            return None
        if game_id == "":
            self.game_id =str(math.floor(random.random()*99999999))
        else:
            self.game_id = game_id
        self.metrics = GameMetrics(self.game_id, dump_path)
        
        self.population = population
        self.performance = performance
        self.myTournament = None
        
        
        self.energy_tracker = {}
        self.data_feed = data_feed
        self.derived_data = derived_data 
        self.multiple_derived_data = multiple_derived_data
        self.game_id = game_id
        self.bulk_energies = {}
        self.bulk_times = {}
        self.game_parms = game_parms
        self.activation_level = game_parms['activation_level']
        self.live_decisions = 0
        self.number_decisions = 0
        
        # --- Data structures for recording decision analysis,data and energies -> for those bots saved ---
        self.decision_raw_data_recorder = {}
        self.bot_decision_tracker = {}
        self.bot_energy_tracker = {}
        
        self.story = {}
        self.label_data = label_data
        
        # --- Reporting JSON ---
        self.static_rpt = {}
        self.dynamic_rpt = {}
        
        # --- Optimisation
        self.hit_idx = {}
        self.op_run = False
        
        # --- DMDB implementation
        self.dmdb_flag = dmdb_flag
        self.dmdb_expression = {}
        self.init_expression_load = init_expression_load
        
       
        
        
    def generation_reset(self):
        # print ("reseting generation data")
        self.performance = None
        self.performance = performance.Performance()
        
        # print ("reseting generation data...done")
        
    def update_dmdb_expression(self,generation_number=0):
        """initialise_dmdb | Initialise output data of dmdb. 
        Loop over DMs in DMDB and build expression structure.
        """
        
        if self.init_expression_load:
            with open('dm/expression_dm.exp', 'rb') as fp:
                self.dmdb_expression = pickle.load(fp)
                tag = list(self.dmdb_expression.keys())[0]
                d = len(self.dmdb_expression[tag])
                
                print (f"length : {d}")
                # for tag, data in self.dmdb_expression.items():
                #     print (len(data))
                
                # print (self.dmdb_expression)
                
                return (1)
        
        # --- Iterate over DMDB
        for dm_id, dm in self.population.dmdb.items():
            if dm_id not in self.dmdb_expression:
                self.dmdb_expression[dm_id] = []
            if len(self.dmdb_expression[dm_id])<1:
                # we have a new dm
                logger_.info(f"Updating expression vector for {dm_id}")
                self.bot_step(dm,dmdb_express_update=True,dm_id = dm_id)
        
        with open('dm/expression_dm.exp', 'wb') as fp:
                pickle.dump(self.dmdb_expression,fp)
        
        return (1)
    
    def play(self):
        """ play | play runs the Ident game main iteration loop. The function iterated
        """
        game_initialised = False
        generation_number = 0
        self.update_dmdb_expression(generation_number = generation_number)
        logger_.trace(self.dmdb_expression)
        
        self.metrics.update_overview(self.game_parms)
                
        
        with Progress() as progress:
            
        # --- Iterate over generations
            for op_run in range(1,self.game_parms['number_generations']):
                # update task bar
                task1 = progress.add_task("[red]Running features (bots)...", total=len(self.population.bots))
                
                # reset generation fitness structure
                self.generation_reset()
                
                # --- Iterate over individuals
                for individual_idx in range(0, len(self.population.bots)):
                    # get bot name and run bot life
                    bot_name = self.population.bots[individual_idx]
                    self.bot_step(self.population.species[bot_name], generation=generation_number)
                    progress.update(task1, advance=1)
                
                
                # --- All individuals have been run -> now evaluate them all
                self.performance.evaluateBots(self.population.species,self.game_parms)
                best_fitness, worst_fitness, winner_id, fitness_data = self.performance.text_output_fitness()
                logger_.info(f'<=== Generation {generation_number} | best [{best_fitness}] | worst [{worst_fitness}] | winning bot [{winner_id}] ===>')
                
                

                # --- We have all the fitness values now, time to *** EVOLVE ***
                saved_ids = self.evolve(generation=generation_number)
                
                
                decision_text = self.performance.showBotDecisions(bot_name=winner_id, verbose=False)
                
                
                self.metrics.update_saved_bots(bot_list = saved_ids, results=self.myTournament.results)
                self.metrics.update_gen({'generation_number' : generation_number, 'best' : best_fitness, 'worst' : worst_fitness, 'winner':winner_id, 'decisions': decision_text, 'saved_bots' : saved_ids, 'results' : self.myTournament.results })
                self.metrics.dump(winners=self.myTournament.winners, generation = generation_number)
                generation_number += 1    
                
    def add_decision(self,listen_start_idx, iter_cnt, bot, iter_start_time, iter_end_time, pressure_id, sr):
        """transcribe _summary_

        :param iter_cnt: _description_
        :type iter_cnt: _type_
        :param bot: _description_
        :type bot: _type_
        :param iter_start_time: _description_
        :type iter_start_time: _type_
        :param iter_end_time: _description_
        :type iter_end_time: _type_
        :param pressure_id: _description_
        :type pressure_id: _type_
        """
        # self.number_decisions += 1
        
        decision_id = 'd_id_' + str(math.floor(random.uniform(0,999999))) + str(math.floor(random.uniform(0,999999))) + str(math.floor(random.uniform(0,999999)))
        # print ("decision made")
        max_memory = bot.GetMemory()
        xr_start_time = iter_start_time - timedelta(milliseconds=max_memory)
        decision_args = {
            'status' : 1,
            'memory' : max_memory,
            'env' : self.game_parms['env'],
            'xr_start' : xr_start_time,
            'iter_start_time' : iter_start_time,
            'iter_end_time' : iter_end_time,
            'action' : 1,
            'type' : "IDent",
            'xr' : -1,
            'epoch' : pressure_id,
            'decision_id' : decision_id,
            'iter_cnt' : iter_cnt
        }
        
        new_decision = IdentDecision(decision_data=decision_args)
        self.performance.add_decision(decision=new_decision, epoch = pressure_id, botID = bot.name)
        
        
        # ================================================
        # query dataset to mark decision
        # ================================================
        

        xr_hits = self.derived_data.query_label_time(xr_start_time, iter_end_time)
        
        # if len(xr_hits)==0:
        #     xr_hits = self.query_label_time(xr_start_time, iter_end_time)

        if len(xr_hits) > 0:
            
            self.live_decisions += 1
            xr_data = xr_hits[0]
            if xr_data['xr'] == True:
                xr = True
                
        else:
            xr = False
        
        # ================================================

        
        decision_args = {
            'status' : 0,
            'memory' : max_memory,
            'xr_start' : xr_start_time,
            'env' : self.game_parms['env'],
            'iter_start_time' : iter_start_time,
            'iter_end_time' : iter_end_time,
            'action' : 0,
            'type' : "IDent",
            'xr' : xr,
            'epoch' : pressure_id,
            'decision_id' : decision_id,
            'iter_cnt' : iter_cnt
        }


        # i_s_i = listen_start_idx - (math.floor(1 * sr))
        # interesting_slice_idx = max(0,(i_s_i))
        # decision_show_data = env_pressure.frequency_ts_np[interesting_slice_idx:slice_end]
        
        # self.decision_raw_data_recorder[decision_id] = decision_show_data
        # if bot.name in self.bot_decision_tracker:
        #     self.bot_decision_tracker[bot.name].append(decision_id)
        # else:
        #     self.bot_decision_tracker[bot.name] = []
        #     self.bot_decision_tracker[bot.name].append(decision_id)
        # record
        close_decision = IdentDecision(decision_data=decision_args)                    
        self.performance.add_decision(decision=close_decision, epoch = pressure_id, botID = bot.name)
    
    def bot_step(self, bot = None, generation=0, dmdb_express_update = False, dm_id=0):
        """bot_step | A single iteration of a bot in the population domain. It iterates over all environment pressures provided to the 
        game.

        :param bot: _description_, defaults to None
        :type bot: _type_, optional
        :param generation: _description_, defaults to 0
        :type generation: int, optional
        """
        
        # full life iter counter
        total_iter_cnt = 0
        # full life time counter
        global_time_counter = 0
        # global time recroder
        global_time_recorder = []
        
        # optimsation mode [not currently active]
        op = False
        interesting_start = 0
        
        # --- Iterate over environment pressures (data snapshot downloaded from data repo)
        for env_pressure in self.data_feed:
            
            # get snapshot id
            pressure_id = env_pressure.meta_data['snapshot_id']
            sr = env_pressure.meta_data['sample_rate']
            # reset counters
            listen_start_idx = 0
            listen_end_idx = 0
            # step size for given delta t of game
            listen_delta_idx = math.floor(self.game_parms['listen_delta_t'] * env_pressure.meta_data['sample_rate'])
            # get total indices in the data vector
            env_pressure_length = env_pressure.frequency_ts_np.shape[0]
            
            # local inter counter
            iter_cnt = 0
            
            # iterate over data vector
            while listen_start_idx < (env_pressure_length - (listen_delta_idx*2)):
                
                # --- START PRE LOGIC HOUSEKEEPING ---
                
                # get start & end slice idx ---
                listen_end_idx = listen_start_idx + listen_delta_idx
                slice_start = listen_start_idx
                slice_end = min(listen_end_idx,env_pressure_length-1)
                
                # get seconds of iter start from start of dataset
                _s = (slice_start / env_pressure.meta_data['sample_rate']) * 1000 # ms 
                # get iteration start time in standard format
                iter_start_time =  env_pressure.start_time + timedelta(milliseconds=_s)
                curr_date_string = iter_start_time.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
                
                # get seconds of iter end from start of dataset
                _s = (slice_end / env_pressure.meta_data['sample_rate']) * 1000
                iter_end_time   =  env_pressure.start_time  + timedelta(milliseconds=_s)
                logger_.trace(f'{iter_start_time} -> {iter_end_time}')
                # logger_.info(f'{global_time_counter} | {iter_end_time}')
                # --- END PRE LOGIC HOUSEKEEPING ---
                
                
                # --- START GAME LOGIC ---
                
                data = {'dmdb_data':self.population.dmdb, 'dmdb_expression_vector' : self.dmdb_expression, 'generation': generation, 'op':op,'bot_id': bot.name,'sim_delta_t':self.game_parms['listen_delta_t'],'global_iter_count':total_iter_cnt ,'timings' : self.game_parms['timings'],'data_index':listen_start_idx - interesting_start, 'sample_rate' : env_pressure.meta_data['sample_rate'] ,'current_data' :  env_pressure.frequency_ts_np.shape[slice_start:slice_end], 'derived_model_data' : self.multiple_derived_data[pressure_id], 'iter_start_time' : iter_start_time, 'iter_end_time' : iter_end_time}
                
                logger_.trace(bot)
                if dmdb_express_update:
                    # update the expression structure for dmdb
                    gene_value = bot.run(data,dmdb_flag = False)
                    # get expression level
                    self.dmdb_expression[dm_id].append(gene_value)
                else:
                    # run genetic expression
                    ex_stat = bot.ExpressDNA(data,dmdb_flag = self.dmdb_flag)
                    # get expression level
                    express_level = bot.GetAvgExpressionValue()
                    # print (express_level)
                    
                    # self.metrics.update_energy(bot.name, generation, energy_data)
                    
                    # get transcription data
                    transcription_data = {
                            'expression_data': bot.GetExpressionData(),
                        }
                    
                    protein_present = bot.transcriptionDNA.run_transcription(bot, dmdb_structure=self.population.dmdb, dmdb_expression_vector=self.dmdb_expression,global_iter_count=total_iter_cnt,transcription_data=transcription_data, dmdb_flag = True)
                    
                    if protein_present:
                        self.add_decision(listen_start_idx,iter_cnt, bot, iter_start_time, iter_end_time, pressure_id, sr)
                        express_level = 0.9
                    # update energy tracker
                    energy_data = {'energy' : express_level, 'time' : curr_date_string}


                # --- END GAME LOGIC ---
                
                
                
                # --- START UPDATE COUNTERS ---
                iter_cnt += 1
                total_iter_cnt += 1
                print (f'total_iter_cnt : {total_iter_cnt}')
                listen_start_idx = listen_end_idx
                global_time_counter = self.game_parms['listen_delta_t'] + global_time_counter
                global_time_recorder.append(global_time_counter)
                # --- END UPDATE COUNTERS ---

    def evolve(self, generation = 0):
    
        """

        Evolution Step 1.
        =============================================
        All the bots have now been evaluated. The bots now compete in a tournament and are ranked. 
        Once ranked, the bot population is regernated. There are a number of ways of doing this. BrahmA
        root structures provide a quick development environment for getting the AI engine off the ground.
        BrahmA provides more sophisticated algorithms and custom algos can also be used. Be sure
        to derive your class definitions from the correct root structures.

        We will use a simple tournament/rank and regeneration algorithm.


        Step 1. Create the Tournament structure and compete.

        """ 
        self.myTournament = None
        self.myTournament = SimpleTournamentRegenerate(generationEval = self.performance.evaluation, population = self.population, dataManager = None)

        """

        Step 2. Tournament Rank

        """
        
        winners = self.myTournament.RankPopulation()
        print (self.myTournament.results)
        self.myTournament.printRankings()
        
        """

        Step 3. Regenerate Population
        Here bots are killed, children are created and the population of bots is 
        regenerated.

        """

        self.myTournament.RegeneratePopulation(dmdb_flag = self.dmdb_flag)
        
        """
        Step 4. The Genetic Shuffle
        Mutation plays a very significant role in evolution by altering gene points within the 
        genome.
        """

        genetic_shuffle = RootMutate(population = self.population, config = self.game_parms )
        genetic_shuffle.mutate(args=self.game_parms, dmdb_flag = self.dmdb_flag)
        
                
        #---generate list of zeros
        
        zeros = self.myTournament.Zeros()
        
        
        #---kill list of zeros
        self.population.KillDeadWood(tags = zeros) 
        #--------------------------------

        #---regenerate population
        self.population.Repopulate(species=self.population.species_name) 
        #--------------------------------
        
    
        # --- Save winning bots ---
        saved_ids = self.population.DMDB_saveBots(self.myTournament.winners, data={'generation' : generation, 'bot_path':self.bot_path})    
        
        return saved_ids
        





"""Population class for evolutionary algorithms
"""
import json, requests
import logging
from tqdm import tqdm as tq
import os, sys, random, pkgutil, math
logging.basicConfig(level=logging.CRITICAL)
#insert root dir into system path for python
#config core
# sys.path.append('../')
# from configCore import *

#import root bot
from marlin_brahma.bots.bot_root import *

# import bots
from custom_bots import *
# from custom_genes import *

import os
from loguru import logger as logger_
# logger_.add('ident_learn.log', format="{level} : {time} : {message}: {process}", level="TRACE")
# # logger_.remove(0)
# logger_.add(sys.stdout, level="INFO")  
import pickle




class Population(object):
    """
    Population is the container object for all bots/indis in the game/optimisation run.
    
    Arguments:
        object {root object} -- object class
    """

    def __init__(self, parms=None, name="vx_dm", gene_args = None, version="1_0_0", dmdb_load = False):
       
        logger_.info("Initialising Population.")
        """
        Create the Population class. This class holds all bots in the game.

        :param parms: optimisation parameters - from config file but passed through for now, defaults to None
        :type parms: key/value pairs, optional
        :param name: name root, defaults to "vx_dm"
        :type name: str, optional
        """

        self.name = name + str(random.randint(10,10000))
        self.parms = parms
        self.feature_version = version
        #bot names in list
        self.bots = []
        #species structure tagged by name
        self.species = {}
        #logger

        #self.logger = GetBrahmaLogger("Population")
        self.gene_args = gene_args
        
        #DM database
        #11-25
        self.dmdb = {}
        self.geneticSelectionDict = {}
        self.dmdb_flag = False
        self.dmdb_load = dmdb_load
        
        
        self.species_name = "undef"
        
    def __exit__(self):
        del self.logger

    #edit march
    #kill bots not making decisions -> 0 fitness
    
    def KillDeadWood(self, tags = []):
        pop_size = len(self.bots)
        dead_size = len(tags)
        logger_.info(f'[ {pop_size} ] bots after generation run. [ {dead_size} ] about to be killed off.')
        
        number = len(tags)
        splice_number = math.floor(number/2)
        # for tag in tags[0:splice_number]:
        for tag in tags:
            if tag in self.species:
                del (self.species[tag])
                self.bots.remove(tag)
        
        pop_size = len(self.bots)
        logger_.info(f'[ {pop_size} ] bots left.')
    #edit march
    def Repopulate(self, species = "BotRoot", args = None):
        population_size = len(self.species)
        delta = self.parms['population_size'] - population_size
        
        args = {
                            'training_set_description' : self.parms['training_set_description'], 
                            'study_focus' : self.parms['study_focus'], 
                            'parent' : "Eve"
                            }
        
        
        for i in range(delta):
            self.CreateBot(species = species, args = args) 
            
        for i in range(2):
            self.CreateBot(species = species, args = args) 
           
        # if dmdb_load == False:
        #     self.save_dmdb()
    

    
    def Populate(self, species = "BotRoot", args = None, dmdb = False):
        
        self.species_name = species
        if dmdb:
            self.dmdb_flag = True
            self.BuildDMDB()
        # logging.debug('Building inside brahma pop ')
        if globals().get("AcousticBot") is not None:
            logger_.info("We have bot creation dust! ")
            
        
        if args == None:
            try:
                args = {
                            'training_set_description' : self.parms['training_set_description'], 
                            'study_focus' : self.parms['study_focus'], 
                            'parent' : "Eve"
                            }
            except:
                logger_.error("Error setting bot args in populution build.")
       
        import pickle
        import os
        
        # if args == None:
        # print ("new bots")
        if species != "tribal":
            # if DEBUG:
            #     print ("Size: " + str(self.parms['population_size']))
            # print ("Size: " + str(self.parms['population_size']))
            population_size = self.parms['population_size']
            # print (self.gene_args)
            logger_.info("Building population.")
            for i in tq(range(0, int(self.parms['population_size']))):
                
                # logger_.info(f'building {i} of {population_size}')
                self.CreateBot(species = species, args = args)   
                # logger.debug(f'built {i} of {population_size}')
                
                #dmdb approach -> randomly select x number from dmdb
            
            logger_.info("Population built.")
            
        else:

            self.CreateBot(species = species, args = args)

        
            
        if args=="Living":
            
            BOT_SAVE_FOLDER = os.path.join(os.path.expanduser('~'), 'dev', 'app', 'tutorial-make-decisions', 'saved', '')
            # load all traders in bin folder
            listOfFiles = os.listdir(BOT_SAVE_FOLDER)
            
            for i in range(0, int(self.parms['population_size'])):
                
                #load saved bots
                filename = random.choice(listOfFiles)
                
                #pop from list
                indx = listOfFiles.index(filename)
                listOfFiles.pop(indx)
                
                try:
                    print (BOT_SAVE_FOLDER + filename)
                    pkl_binary = open(BOT_SAVE_FOLDER + filename , 'rb')
                   
                except:
                    print ("warning, bot not found")
                    exit()
                    continue
            
                try:
                    _bot = pickle.load(pkl_binary)
                    self.ResetTrader(_bot)
                    self.bots.append(_bot.name)
                    self.species[_bot.name] = _bot
                except Exception as e:
                    print (e)
                    
                
                # print (_bot.name)
                # print (_bot)
                
        
       
                    
    def ResetTrader(self,  trader):
        for dnaTag, dna in trader.dNA.items():
            dna.expressionTable = {}
            

    def test(self):
        pass

    
    # ----- NEW FRAMEWORK FOR OPTIMISATION ------
    def save_dmdb(self):
        with open('dm/saved_dmdb.dm', 'wb') as fp:
            pickle.dump(self.dmdb,fp)

    def BuildDMDB(self, args=None):
        
        if self.dmdb_load:
            with open('dm/saved_dmdb.dm', 'rb') as fp:
                self.dmdb = pickle.load(fp)
                
                
            
            return (1)
        
        population_size = self.parms['dmdb_size']
        logger_.info("Building [DMDB].")
        for i in tq(range(0, int(self.parms['dmdb_size']))):
                
            # logger_.debug(f'building {i} of {population_size}')
            self.CreateDMDBInd(args = args)   
            logger_.debug(f'built {i} of {population_size}')
    
        return (1)
    
    def CreateDMDBInd(self, args=None):
        
        #self.logger.info("Building Normal Genome - Gene count: {0}".format(self.genomeSize))
        self.BuildGeneSelectionDict()
        logger_.debug(self.BuildGeneSelectionDict)
        counter = {}
        #select
        valid_gene = False
        while not valid_gene:
            geneType = self.SelectGeneFromSelectionDict()
            logger_.debug(geneType)
            # print (self.gene_args)
            # print (geneType)
            if geneType in counter:
                if counter[geneType] < self.gene_args[geneType]:
                    valid_gene = True
                    counter[geneType] += 1
            else:
                if self.gene_args[geneType] != 0:
                    counter[geneType] = 1
                    valid_gene = True
                
            
            
            logger_.debug("Normal gene selected: {0}".format(geneType))

            
            geneStructureTmp = self.BuildGeneStructure(geneType,  self.geneticSelectionDict, gene_args=self.gene_args)
            geneStructureTmp.i_D = f'{geneStructureTmp.i_D}_dmdb'
            self.dmdb[geneStructureTmp.i_D] = geneStructureTmp
            #kill tmp structure
            geneStructureTmp = None
            

        return
 
    def BuildGeneSelectionDict(self):
        """Build gene type selection lists
        """


        import os
        #local genes loaded {todo: custom genes}
        #geneFileList = os.listdir(BRAHMA_GENETIC_DATA_FOLDER_LCL)
        #edit
        
        genetic_material_folder  = os.getenv('GENETIC_DATA_FOLDER_USR')
        logger_.debug(f'{genetic_material_folder}')
        geneFileList = os.listdir(genetic_material_folder)
        logger_.debug(f'{geneFileList}')
        for fileName in geneFileList:
            
            if fileName.endswith('.py'):
                
                fnArray = fileName.split('.')
                geneType = fnArray[0]
                if geneType != "dnaroot":
                    
                    prefix = geneType.split('_')
                    prefixStr = prefix[0]
                    
                    if prefixStr == "g":
                        geneName = prefix[1]
                        
                        logger_.debug("Adding Gene to Selection: {0}".format(geneName))
                        self.geneticSelectionDict[geneName] = eval(f'{geneName}')
            
    def SelectGeneFromSelectionDict(self):
        """Select a genetype from the list
        and return to caller
        """
        import random
        try:
            #self.logger.debug("Trying to grab a random gene tag.")
            
            randomGeneTag = random.choice(list(self.geneticSelectionDict))
            #self.logger.debug("Random gene tag: {0}".format(randomGeneTag))
            
            return randomGeneTag

        except Exception as e:
            
            #self.logger.critical("Error selecting gene")
            exit(-1)
        return -1
 
    def BuildGeneStructure(self, gene_type = "", classCollection=None, gene_args = {}):
        if gene_type != "":
            pass
        else:
            #self.logger.critical("No genetic material found for gene_type: {0}".format(gene_type))
            exit("No genetic material")

        logger_.debug(f"Building gene : {gene_type}")
        
        logger_.debug(f"Building gene : {classCollection}")
        
        tmpGene = None
        #self.logger.debug("attempting to create gene from new selection approach: {0}".format(gene_type))
        try:
            #print (gene_type)
            #print (self.geneticSelectionDict)
            #tmpGene = self.geneticSelectionDict[gene_type]("test")
            #'test' being passed as env changed  feb '21
            # print ("***")
            # print (classCollection)
            
            tmpGene = classCollection[gene_type](self.parms['env'], gene_args=gene_args)
            # print ("---")
            # print (f'temp : {tmpGene}')
        except Exception as e:
            #self.logger.critical("Attempting to create gene from new selection approach: FAIL")
            logger_.error(e)
            exit(0)

        return tmpGene

    def ShowDMDB(self):
        print (self.dmdb)
        for dm_id, dm in self.dmdb.items():
            desc = dm.__str__() 
            logger_.info(f'{dm_id} : {desc}')
    

    # ----- NEW FRAMEWORK FOR OPTIMISATION ------

    def CreateBot(self, species = None, args = None):
        
        self.botStr = {}
        self.botStr[species] = eval(species)
        
        logger_.trace(f'creating bot')
        #build the bot here - general
        
        bot_tmp = self.botStr[species](self.parms["env"], myspecies = species, myargs = args, version=self.feature_version)
        
        bot_tmp.BuildBot(parms=self.parms, gene_args = self.gene_args, dmdb_flag = self.dmdb_flag, dmdb_ids = list(self.dmdb.keys()))
        
        # bot_tmp.printAll()
        self.bots.append(bot_tmp.name)
        self.species[bot_tmp.name] = bot_tmp
        
       

    def Show(self):
        
        for key, value in self.species.items():
            # print (value)
            value.printAll()

            
    def save_bots(self):
        for bot_id, bot in self.species.items():
            
            #serialise bot to local file.
            print (f'saving ... {bot_id}')
            bot.save()
            # self.recordWinningBot(bot)
            
    
    def DMDB_saveBots(self,bot_list = [], data={}):
        """DMDB_saveBots    Save list of bots. We need to build a structure which can be used live. We need to swap the string ID of a gene (DM)
        for a real gene structure that can be used live. This is done by iterating over the genome in a DNA strand and swapping a DM ID with a DM strucure.
        The structure is taken from the DMDB structure.

        :param bot_list: _description_, defaults to []
        :type bot_list: list, optional
        """
        
        new_names = []
        
        suffix = ""
        if 'generation' in data:
            suffix = data['generation']
            
        
        # iterate over bot ids
        for bot_id in bot_list:
            # create a deep copy of the active bot
            logger_.info(f'Saving {bot_id}')
            bot_tmp = copy.deepcopy(self.species[bot_id])
            bot_tmp.name = self.species[bot_id].name + "_" + str(suffix)
            # get active DNA & genome id tag
            dNA_tag = list(bot_tmp.dNA.keys())[0]
            genome_tag = list(bot_tmp.dNA[dNA_tag].genome.keys())[0]
            
            # get list of genes in genome
            _genome_ = bot_tmp.dNA[dNA_tag].genome[genome_tag].genome
            
            # iterate over genome and swap gene id for structure
            _new_genome = {}
            for g_tag, dm_id in _genome_.items():
                dm_str = self.dmdb[dm_id]
                _new_genome[g_tag] = dm_str
            
            bot_tmp.dNA[dNA_tag].genome[genome_tag].genome = _new_genome
            # bot_tmp.printAll()
            bot_tmp.Save(save_folder = data['bot_path'])
            new_names.append(bot_tmp.name)
            
            
        return new_names
    

    
    def recordWinningBot(self, bot):
        """Record optimisation winners to db

        Arguments:
            botlist {[type]} -- [description]
        """
        #logger.info('Winning bots recorded')
        botStr = bot
        botID = bot.name
        #--build data for posting
        dataSend = {}
        dataSend["action"] = "record_bot"
        dataSend["user"] = self.parms['uid']
        dataSend["botID"] = botID
        dataSend["botStructure"] = botStr.printStr()
        dataSend["market"] = self.parms['env']
        dataSend["direction"] = ""
        dataSend["optimisationID"] = ""
        dataSend["parent"] = "Eve"
        dataSend["scope"] = "global"
       
        
        dataSendJSON = json.dumps(dataSend)
      
        #--post data
        try:
            API_ENDPOINT = "https://www.vixencapital.com/api/optimisation/"
            try:
                r = requests.post(url = API_ENDPOINT, data = dataSendJSON)
                response = r.text
            except:
                print("request error")

        except:
            #logger.critical('Error recording winning bot')
            pass


        #print (response)
    


if __name__ == "__main__":
    pass
    '''
    pop = Population(parms={'PopulationSize' : '100'}, name = "demo")
    pop.populate('trader')
    '''

